#define WIDTH	320
#define HEIGHT	240

#define BASE	0x00200000

.macro	thumb
	add	r1, pc, #1
	bx	r1
	.thumb
.endm

.macro	endthumb
	adr	r1, 9f 
	bx	r1
	.arm
	.align 2
	9:
.endm

	.text

	.align	2
start:

	.word	0x464c457f			// e_ident
dev_mem_filename:
	.ascii "/dev/mem"
	.word	0x00000000
	.word	0x00280002			// e_machine | e_type
	.word	0x00000000			//(e_version)
	.word	entry-start+BASE	// e_entry
	.word	prog_header-start	// e_phoff
prog_header:
	.word	0x00000001			//(e_shoff)						// p_type
	.word	0x00000000			//(e_flags)						// p_offset
	.word	BASE				// e_phentsize |(e_ehsize)		// p_vaddr
	.word	0x00000001			//(e_shentsize)| e_phnum		//(p_paddr)
	.word	end-start			//(e_shstrndx) |(e_shnum)		// p_filesz
	.word	end-start											//(p_memsz)
mmap_params:
	.word	0x00000002			// mmap_addr					// p_flags (1<<1)
	.word	0x00020000			// mmap_size					//(p_align)
	.word	3					// mmap_prot = PROT_READ | PROT_WRITE
	.word	1					// mmap_flags = MAP_SHARED
mmap_params_fd:
	.word	3					// mmap_fd
mmap_params_offset:
	.word	0x03101000			// mmap_offset

	.align	2
	.global	entry
entry:
	// open /dev/mem
	adr		r0, dev_mem_filename
	mov		r1, #2				// O_RDWR
	swi		#0x900005			// sys_open

	str		r0, mmap_params_fd

	// mmap the vram
	adr		r0, mmap_params
	swi		#0x90005a			// old_mmap
	mov		r11, r0				// r11 = vram addr

	// mmap the gpio regs
	adr		r0, mmap_params
	mov		r2, #0xc0000000
	str		r2, mmap_params_offset
	swi		#0x90005a			// old_mmap
	add		r12, r0, #0x1180	// r12 = gpio addr + 0x1180


thumb

	adr		r1, reglist
1:
	ldmia	r1!, { r2 }
	lsr		r7, r2, #16
	beq		1f
	strh	r2, [r0, r7]
	strh	r2, [r0, r7]
	b		1b
1:



init:

	// allocate working buffers in cached memory (on the stack)
	// r6 = current buffer, r7 = offset to next buffer
	ldr		r0, height_times_width
	neg		r7, r0
	mov		r6, sp
	add		r6, r6, r7

	// init the rng
	adr		r5, lcrng_params
	ldmia	r5!, { r3, r4, r5 }

1:
	// crank the rng
	mul		r3, r3, r4
	add		r3, r3, r5
	
	// just use the top bit
	lsr		r1, r3, #31

	sub		r0, r0, #1
	strb	r1, [r6, r0]
	bne		1b


	// store random seed
	adr		r5, lcrng_params
	str		r3, [r5]


main_loop:
	
	bl		vsync

	// read input - we need to do this before render so we can effectively double-vsync
	//mov		r2, r12			// already set by vsync
	ldrh	r5, [r2, #4]		// read register C
	asr		r5, r5, #8			// shift so we can easily test the right bits

	mov		r4, #0x10
	tst		r5, r4				// A = slow
	bne		noslow

slow:
	bl		vsync
	sub		r4, r4, #1
	bne		slow
	
noslow:

	// render
	ldr		r0, height_times_width
	mov		r4, r11
1:
	sub		r0, r0, #4
	ldr		r1, [r6, r0]
	str		r1, [r4, r0]
	bne		1b


	// more input processing - r5 is still valid
	mov		r4, #0x40
	tst		r5, r4				// X = pause
	beq		main_loop

	mov		r4, #0x20
	tst		r5, r4				// B = reset
	beq		init

	mov		r4, #0x01
	tst		r5, r4				// start = quit
	bne		1f
	adr		r1, exit
	bx		r1
1:

	// update next generation buffer
	ldr		r5, height_minus_two_times_width
	ldr		r4, width
	add		r3, r6, r4			// note that r3 points at pixel to the left of current - we can't add #WIDTH+1 :(
1:
	ldrb	r0, [r3]
	ldrb	r1, [r3, #2]
	add		r0, r0, r1
	sub		r2, r3, r4			// row above
	ldrb	r1, [r2]
	add		r0, r0, r1
	ldrb	r1, [r2, #1]
	add		r0, r0, r1
	ldrb	r1, [r2, #2]
	add		r0, r0, r1
	add		r2, r3, r4			// row below
	ldrb	r1, [r2]
	add		r0, r0, r1
	ldrb	r1, [r2, #1]
	add		r0, r0, r1
	ldrb	r1, [r2, #2]
	add		r0, r0, r1
	sub		r0, r0, #2
	bne		not_2_neighbours
	ldrb	r0, [r3, #1]
not_2_neighbours:
	cmp		r0, #1
	beq		writepix
	mov		r0, #0
writepix:
	add		r3, r3, #1
	strb	r0, [r3, r7]		// write pixel in next buffer

	sub		r5, r5, #1
	bne		1b


	// flip buffers
	add		r6, r6, r7
	neg		r7, r7


	// sync vertical edges
	mov		r0, #HEIGHT
	sub		r1, r6, #2		// r1 = 2 pixels off left edge
	ldr		r3, width
1:
	sub		r0, r0, #1
	blt		1f
	add		r2, r1, r3		// r2 = 2nd pixel in from right edge
	ldrb	r4, [r2]
	ldrb	r5, [r1, #3]
	strb	r4, [r1, #2]
	strb	r5, [r2, #1]
	mov		r1, r2
	b		1b
1:

	// sync horizontal edges
	add		r2, r2, #2				// adjust leftover r2 value from above
	mov		r1, r6					// top
	add		r4, r1, r3				// second from top
	sub		r2, r2, r3				// bottom
	sub		r5, r2, r3				// second from bottom
1:
	sub		r3, r3, #4
	ldr		r0, [r4, r3]
	str		r0, [r2, r3]
	ldr		r0, [r5, r3]
	str		r0, [r1, r3]
	bne		1b


	b		main_loop


vsync:
	mov		r2, r12
	mov		r0, #0x10
1:
	ldrh	r1, [r2, #2]
	tst		r1, r0
	bne		1b
1:
	ldrh	r1, [r2, #2]
	tst		r1, r0
	beq		1b
	mov		pc, lr


	.align	2
reglist:
	.word	0x28da02ab		// store 0x02ab in gpio 0x28da
	.word	0x290c0140		// store 320    in gpio 0x290c
	.word	0x29580000		// set palette starting at entry 0
	.word	0x295affff
	.word	0x295a0000

lcrng_params:
	.word	0
	.word	0x0019660d			// lcrng_A
	.word	0x3c6ef35f			// lcrng_B

width:
	.word	WIDTH

height_times_width:
	.word	HEIGHT*WIDTH

height_minus_two_times_width:
	.word	(HEIGHT-2)*WIDTH

exit:
	.arm
	swi		#0x900001

end:	// mark end of program for header calculations

